%% tclp/tclp__type_order.pl
%% rewritten version
%% Copyright (C) 2003 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                   TCLP                                    %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__type_order2.pl,v 1.2 2003/08/29 15:55:09 manu Exp $

:- module(tclp__type_order, [ add_type/1, 
                              add_type/2,
                              add_order/2,
                              build_order/0,
                              tclp__glb/3,
                              tclp__lub/3,
                              tclp__break/2,
                              tclp__ordered_types/2,
			      tclp__elim_down_i/3,
			      tclp__elim_up_i/3,
                              type_exists/1,
                              init_order/0
                              ]).


:- use_module(tclp(tclp__utils)).
:- use_module(tclp(tclp__arity)).
:- use_module(tclp(tclp__errors)).
:- use_module(tclp(tclp__arrays)).
:- use_module(library(lists)).
:- use_module(library(chr)).
:- use_module(tclp(tclp__def)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% initialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
init_order :-
	init_counter(tclp_type_nb).

build_order :-
	build_matrixes.

:- init_order.

reinit :-
	retractall(type_id(_,_)),
	retractall(id_type(_,_)),
	retractall(type_order_dec(_,_)),
	retractall(type_dec(_,_)),
	init_order.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Memoization of types
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%
%% dynamic predicates used for storing declared types and order declarations

:- dynamic type_id/2, id_type/2, type_order_dec/2, type_dec/2.

add_type(T) :-
        functor(T,C,N), add_type(C,N).

add_type(C,_) :-
	type_dec(C,_),!.
add_type(C,N) :-
        arity(C,N),
        incr_counter(tclp_type_nb,_,TypeNb),
        assert(type_id(C,TypeNb)),
        assert(id_type(TypeNb,C)),
	assert(type_dec(C,N)),
	force_order(C).

add_order(T1,T2) :-
        unifyToArity(T1), unifyToArity(T2),
        functor(T1,C1,_), functor(T2,C2,_),
        assert(type_order_dec(C1,C2)).

force_order(T) :-
	builtin_type(top, Top),add_type(Top),
	functor(T,C,N), add_order(T,Top),
	( T == Top -> true ;
	    builtin_type(term, Term),add_type(Term),
	    functor(T,C,N), add_order(T,Term) ).
	
%% converts a list of ids into a list of types
map_id_type([],[]).
map_id_type([Id|Ids],[Type|Types]) :-
        id_type(Id,Type),
        map_id_type(Ids,Types).

%% converts a list of types into a list of ids
map_type_id([],[]).
map_type_id([Type|Types],[Id|Ids]) :-
        typeid_(Type,Id),
        map_type_id(Types,Ids).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Finding extremas and completing the type structure with extra that have no
%% type parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%
%% finding tops and bottoms
get_all_tops(Tops) :-
        findall(Top,
                (type_id(Top,_), \+ (type_order_dec(Top,C), C\= Top)),
                Tops).

get_all_bottoms(Bots) :-
        findall(Bot,
                (type_id(Bot,_), \+ (type_order_dec(C,Bot), C\= Bot)),
                Bots).

%%%%%%%%%
%% completing with empty arity types

complete_tops([]).
complete_tops([X|Rest]) :-
        arity(X,N),
        ( N=0 -> true
        ; atom_concat('$empty_',X,EX), add_type(EX,0),
            functor(T,X,N), add_order(T,EX)),
        complete_tops(Rest).

complete_bottoms([]).
complete_bottoms([X|Rest]) :-
        arity(X,N),
        ( N=0 -> true
        ; atom_concat('$empty_',X,EX), add_type(EX,0),
            functor(B,X,N), add_order(EX,B)),
        complete_bottoms(Rest).

complete_order :-
        get_all_tops(Tops), complete_tops(Tops),
        get_all_bottoms(Bottoms), complete_bottoms(Bottoms),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Building type matrix
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% lookup in the type matrix
ordered_in_matrix(Matrix, Id, Id) :-
        matrix_element_at(Matrix,Id,Id,Answer),
        Answer = y,!.
ordered_in_matrix(Matrix, Id1, Id2) :-
        matrix_element_at(Matrix,Id1,Id2,Answer),
        Answer == y.

%% add an inequality in the matrix
add_leq_in_matrix(Matrix, Id1, Id2) :-
        matrix_element_at(Matrix,Id1,Id2,y).

%% build a matrix of correct size
create_order_matrix(Matrix) :-
        current_counter_value(tclp_type_nb, Size),
        create_matrix(Size,Size, Matrix),
	init_order_matrix(Matrix),
	compute_matrix_transitive_closure(Matrix,Size).

%% fill the matrix with type declarations
init_order_matrix(Matrix) :-
        findall(Id1-Id2,
                (type_order_dec(C1,C2), type_id(C1,Id1), type_id(C2,Id2)),
                Decs),
        init_order_matrix2(Matrix, Decs).

init_order_matrix2(_,[]).
init_order_matrix2(Matrix,[Id1-Id2|Rest]) :-
        add_leq_in_matrix(Matrix,Id1,Id2),
        init_order_matrix2(Matrix,Rest).

%% Compute transitive closure using Warshall's algorithm
compute_matrix_transitive_closure(Matrix,Size) :-
        warshall1(Matrix,Size,1).

warshall1(_,Size,N) :-
        Size < N,!.
warshall1(Matrix,Size,N) :-
        warshall2(Matrix,Size,N,1),
        N1 is N+1,
        warshall1(Matrix,Size,N1).
warshall2(_Matrix,Size,_N,I) :-
        Size < I,!.
warshall2(Matrix,Size,N,I) :-
        warshall3(Matrix,Size,N,I,1),
        I1 is I+1,
        warshall2(Matrix,Size,N,I1).
warshall3(_Matrix,Size,_N,_I,J) :-
        Size < J,!.
warshall3(Matrix,Size,N,I,J) :-
        (
          ordered_in_matrix(Matrix,I,N),
          ordered_in_matrix(Matrix,N,J)
        ->
          add_leq_in_matrix(Matrix,I,J)
        ;
          true
        ),
        J1 is J+1,
        warshall3(Matrix,Size,N,I,J1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Reverse a matrix
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

reverse_matrix(Matrix, Reversed) :-
        current_counter_value(tclp_type_nb, Size),
        create_matrix(Size,Size, Reversed),
        reverse_matrix2(Matrix,Reversed,1,1,Size).
reverse_matrix2(_,_,N,_,Size) :-
        N > Size,!.
reverse_matrix2(Matrix,Reversed,N,M,Size) :-
        M > Size,!,
        N1 is N+1,
        reverse_matrix2(Matrix,Reversed,N1,1,Size).
reverse_matrix2(Matrix,Reversed,N,M,Size) :-
        ( ordered_in_matrix(Matrix,N,M) ->
            add_leq_in_matrix(Matrix,M,N) ; true),
        M1 is M+1,
        reverse_matrix2(Matrix,Reversed,N,M1,Size).
            

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Building the glb/lub matrix
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% true if Id is smaller than all ids of the list
all_smaller(_Matrix,_Id,[]).
all_smaller(Matrix,Id,[Id1|Rest]) :-
        ordered_in_matrix(Matrix,Id,Id1),
        all_smaller(Matrix,Id,Rest).

%% find all types constructors (by ids) smaller than the given list of
%% type constructor
find_smaller(Matrix,Ids,Smaller) :-
        current_counter_value(tclp_type_nb,Size),
        find_smaller2(Matrix, Ids, 1, Size, Smaller).
find_smaller2(_,_, N, Size, []) :-
        N > Size,!.
find_smaller2(Matrix, Ids, N, Size, Smaller) :-
        ( all_smaller(Matrix,N,Ids) ->
            Smaller = [ N | Rest ] ;
            Smaller = Rest ),
        N1 is N+1, !,
        find_smaller2(Matrix, Ids, N1, Size, Rest).

%% find the maximal elements of the given list
find_maximals(Matrix,IdList, Maximals) :-
        findall(Id,
                ( member(Id,IdList),
                    \+ ( member(Id2, IdList), Id2 \= Id,
                           ordered_in_matrix(Matrix,Id,Id2) )
                ),
                Maximals).

%% find the greatest lower bound of two Ids
find_glb(Matrix, Id1, Id2, Id1, _) :-
        ordered_in_matrix(Matrix,Id1,Id2),!.
find_glb(Matrix, Id1, Id2, Id2, _) :-
        ordered_in_matrix(Matrix,Id2,Id1),!.
find_glb(Matrix, Id1, Id2, GLBId, GLBorLUB) :-
        find_smaller(Matrix,[Id1,Id2],Smaller),
        find_maximals(Matrix,Smaller,Maximals),!,
        length(Maximals,N),
        (N > 1 ->
            map_id_type([Id1,Id2|Maximals],[Type1,Type2|Types]),
            print_error(['Multiple ',GLBorLUB,' candidates for ',
                         Type1,' and ',Type2,': ',Types])
        ;
            Maximals=[GLBId]
        ).

%% creates a glb matrix from an order matrix
make_glb_matrix(Matrix,GLBMatrix) :-
        current_counter_value(tclp_type_nb,Size),
        create_matrix(Size,Size,GLBMatrix),
        fill_glb_matrix(Matrix,GLBMatrix,1,1,Size,'GLB').
%% fills the glb matrix
fill_glb_matrix(_Matrix,_GLBMatrix,N,_M,Size,_GLBorLUB) :-
	N > Size,!.
fill_glb_matrix(Matrix,GLBMatrix,N,M,Size,GLBorLUB) :-
        M > Size, N1 is N+1,!,
        fill_glb_matrix(Matrix,GLBMatrix,N1,1,Size,GLBorLUB).
fill_glb_matrix(Matrix,GLBMatrix,N,M,Size,GLBorLUB) :-
	(find_glb(Matrix, N, M, GLB, GLBorLUB) ->
            matrix_element_at(GLBMatrix,N,M,GLB)
        ;
            true
        ),
        M1 is M+1,
        fill_glb_matrix(Matrix,GLBMatrix,N,M1,Size,GLBorLUB).

%% gets the id of the glb of two ids
get_glb(GLBMatrix, Id1, Id2, GLBId) :-
        matrix_element_at(GLBMatrix, Id1, Id2, GLBId),
        nonvar(GLBId).

%% return the glb of a list
list_glb(GLBMatrix, [ Id | Ids ], GLBId) :-
        list_glb2(GLBMatrix, Id, Ids, GLBId).
list_glb2(_,Id,[],Id).
list_glb2(GLBMatrix, Id, [ Id2 | Ids ], GLBId) :-
        get_glb(GLBMatrix, Id, Id2, NewId),
        list_glb2(GLBMatrix, NewId, Ids, GLBId).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Upper/Lower bound of a type w.r.t. some argument id
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% true if the given type constructor does not have X as the arity of one of
%% its arguments.
%% X is a distiguished variable !
does_not_have_arity(Id,X) :-
        id_type(Id,Cons),
        arity(Cons,N), functor(Type,Cons,N),
        unifyToArity(Type),
        Type =.. [Cons|Args],!,
        \+ true_member(X,Args).

%% use this last function to filter a list
dont_have_arity([],_,[]).
dont_have_arity([Id|Rest],X,Result) :-
        ( does_not_have_arity(Id,X) ->
            Result = [Id|R] ;
            Result = R ),
        dont_have_arity(Rest,X,R).

%% find all ids smaller than a given id but with no argument corresponding
find_lb_wrt_i(Matrix,LUBMatrix,Id,X,Result) :-
        find_smaller(Matrix,[Id],Raw),
        dont_have_arity(Raw,X,Filtered),!,
        list_glb(LUBMatrix, Filtered, Result).

%% builds a term for lookup of the id of a lb wrt elimination of an argument
make_lb_wrt_i_term(Matrix,LUBMatrix,Id,Term) :-
        id_type(Id,Cons),
        arity(Cons,N),
        functor(Term,Cons,N),
        Term =.. [Cons|Args],
        make_lb_wrt_i_term2(Matrix,LUBMatrix,Id,Cons,Args,1).
make_lb_wrt_i_term2(_,_,_,_,[],_).
make_lb_wrt_i_term2(Matrix,LUBMatrix,Id,Cons,[Arg|Args],N) :-
        arity(X,Cons,N),
        find_lb_wrt_i(Matrix,LUBMatrix,Id,X,Arg),!,
	N1 is N+1,
        make_lb_wrt_i_term2(Matrix,LUBMatrix,Id,Cons,Args,N1).
make_lb_wrt_i_term2(Matrix,LUBMatrix,Id,Cons,[_Arg|Args],N) :-
	N1 is N+1,
        make_lb_wrt_i_term2(Matrix,LUBMatrix,Id,Cons,Args,N1).

%% builds an array with the id of defaulting types
make_lb_wrt_i_array(Matrix, LUBMatrix, Array) :-
        current_counter_value(tclp_type_nb,Size),
        create_array(Size,Array),
        fill_lb_wrt_i_array(Matrix, LUBMatrix, Array, 1, Size).
fill_lb_wrt_i_array(_,_,_,N,Size) :-
        N > Size ,!.
fill_lb_wrt_i_array(Matrix, LUBMatrix, Array, N, Size) :-
        array_element_at(Array,N,E),
        make_lb_wrt_i_term(Matrix,LUBMatrix,N,E),
        N1 is N+1,
        fill_lb_wrt_i_array(Matrix, LUBMatrix, Array, N1, Size).

get_b_wrt_i(Array,Id,I,NId) :-
        array_element_at(Array,Id,E),
        arg(I,E,NId),!, nonvar(NId).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Generate Clauses for type glb
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

make_glb_clause_with_args(GLBMatrix, Id1, Id2, Type1, Type2, TypeGLB, Goal) :-
        get_glb(GLBMatrix, Id1, Id2, GLBId),
        id_type(Id1,Cons1), id_type(Id2,Cons2), id_type(GLBId, ConsGLB),
        arity(Cons1, N1), arity(Cons2, N2), arity(ConsGLB, NGLB),
        functor(Type1, Cons1, N1),
        functor(Type2, Cons2, N2),
        functor(TypeGLB, ConsGLB, NGLB),
        fold(Type1, Type2, TypeGLB,
             tclp__type_order:add_eq, tclp__type_order:add_vGLB, !, Goal),!
	;
	Goal=fail.

add_eq(A, B, Accu, ( Accu , A=B)).
add_vGLB(A,B,GLB, Accu, ( Accu , tclp__clp_types:tclp__vGLB(A,B,GLB))).

make_glb_clause_matrix(GLBMatrix, GLBClauseMatrix) :-
	current_counter_value(tclp_type_nb,Size),
	create_matrix(Size,Size,GLBClauseMatrix),
	init_glb_clause_matrix(GLBMatrix, GLBClauseMatrix, 1, 1, Size).
init_glb_clause_matrix(_,_,N,_,Size) :- N > Size,!.
init_glb_clause_matrix(GLBMatrix, GLBClauseMatrix, N, M, Size) :-
	M > Size, !, N1 is N+1,
	init_glb_clause_matrix(GLBMatrix, GLBClauseMatrix, N1, 1, Size).
init_glb_clause_matrix(GLBMatrix, GLBClauseMatrix, N, M, Size) :-
	make_glb_clause_with_args(GLBMatrix, N, M, T1, T2, TGLB, Goal),
	matrix_element_at(GLBClauseMatrix, N, M, g(T1, T2, TGLB, Goal)),
	M1 is M+1,
	init_glb_clause_matrix(GLBMatrix, GLBClauseMatrix, N, M1, Size).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Generate Clauses for lubs
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

make_lub_clause_with_args(LUBMatrix, Id1, Id2, Type1, Type2, TypeLUB, Goal) :-
        get_glb(LUBMatrix, Id1, Id2, LUBId)
        ->
        id_type(Id1,Cons1), id_type(Id2,Cons2), id_type(LUBId, ConsLUB),
        arity(Cons1, N1), arity(Cons2, N2), arity(ConsLUB, NLUB),
        functor(Type1, Cons1, N1),
        functor(Type2, Cons2, N2),
        functor(TypeLUB, ConsLUB, NLUB),
        fold(Type1, Type2, TypeLUB,
             tclp__type_order:add_eq, tclp__type_order:add_vLUB, !, Goal)
        ;
        Goal=fail.

add_vLUB(A,B,LUB, Accu, ( Accu , tclp__clp_types:tclp__vLUB(A,B,LUB))).

make_lub_clause_matrix(LUBMatrix, LUBClauseMatrix) :-
	current_counter_value(tclp_type_nb,Size),
	create_matrix(Size,Size,LUBClauseMatrix),
	init_lub_clause_matrix(LUBMatrix, LUBClauseMatrix, 1, 1, Size).
init_lub_clause_matrix(_,_,N,_,Size) :- N > Size,!.
init_lub_clause_matrix(LUBMatrix, LUBClauseMatrix, N, M, Size) :-
	M > Size, !, N1 is N+1,
	init_lub_clause_matrix(LUBMatrix, LUBClauseMatrix, N1, 1, Size).
init_lub_clause_matrix(LUBMatrix, LUBClauseMatrix, N, M, Size) :-
	make_lub_clause_with_args(LUBMatrix, N, M, T1, T2, TLUB, Goal),
	matrix_element_at(LUBClauseMatrix, N, M, g(T1, T2, TLUB, Goal)),
	M1 is M+1,
	init_lub_clause_matrix(LUBMatrix, LUBClauseMatrix, N, M1, Size).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Generate clauses for break
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

make_break_clause_with_args(Matrix, Id1, Id2, Type1, Type2, Goal) :-
        ordered_in_matrix(Matrix,Id1,Id2)
        -> 
        id_type(Id1,Tau1), id_type(Id2,Tau2),
        arity(Tau1, N1), arity(Tau2, N2), 
        functor(Type1, Tau1, N1), functor(Type2, Tau2, N2),
        fold(Type1, Type2, tclp__type_order:add_leq, !, Goal)
        ;
        Goal=fail.

add_leq(A,B,Accu, ( Accu , tclp__clp_types:':<'(A,B) , verbose(debug_clp_types,['break ',A,' :< ',B]))).

make_break_clause_matrix(Matrix, BreaKMatrix) :-
	current_counter_value(tclp_type_nb,Size),
	create_matrix(Size,Size,BreaKMatrix),
	init_break_clause_matrix(Matrix, BreaKMatrix, 1, 1, Size).
init_break_clause_matrix(_,_,N,_,Size) :- N > Size,!.
init_break_clause_matrix(Matrix, BreaKMatrix, N, M, Size) :-
	M > Size,!, N1 is N+1,
	init_break_clause_matrix(Matrix, BreaKMatrix, N1, 1, Size).
init_break_clause_matrix(Matrix, BreaKMatrix, N, M, Size) :-
	make_break_clause_with_args(Matrix, N, M, T1, T2, Goal),
	matrix_element_at(BreaKMatrix, N, M, g(T1, T2, Goal)),
	M1 is M+1,
	init_break_clause_matrix(Matrix, BreaKMatrix, N, M1, Size).
	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Generate clauses for ordered_types
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

make_ordered_types_clause_with_args(Matrix, Id1, Id2, Type1, Type2, Goal) :-
        ordered_in_matrix(Matrix,Id1,Id2)
        -> 
        id_type(Id1,Tau1), id_type(Id2,Tau2),
        arity(Tau1, N1), arity(Tau2, N2), 
        functor(Type1, Tau1, N1), functor(Type2, Tau2, N2),
        fold(Type1, Type2, tclp__type_order:add_ordered_vars, !, Goal)
        ;
        Goal=fail.

add_ordered_vars(A,B,Accu, ( Accu, tclp__clp_types:ordered_vars(A,B) )).

make_ordered_clause_matrix(Matrix, OrderedMatrix) :-
	current_counter_value(tclp_type_nb,Size),
	create_matrix(Size,Size,OrderedMatrix),
	init_ordered_clause_matrix(Matrix, OrderedMatrix, 1, 1, Size).
init_ordered_clause_matrix(_,_,N,_,Size) :- N > Size,!.
init_ordered_clause_matrix(Matrix, OrderedMatrix, N, M, Size) :-
	M > Size,!, N1 is N+1,
	init_ordered_clause_matrix(Matrix, OrderedMatrix, N1, 1, Size).
init_ordered_clause_matrix(Matrix, OrderedMatrix, N, M, Size) :-
	make_ordered_types_clause_with_args(Matrix, N, M, T1, T2, Goal),
	matrix_element_at(OrderedMatrix, N, M, g(T1, T2, Goal)),
	M1 is M+1,
	init_ordered_clause_matrix(Matrix, OrderedMatrix, N, M1, Size).
	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Generate clauses for arg deletion
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

make_arg_delete_clause_with_args(Array, Id, I, Type, NewType, Goal) :-
        get_b_wrt_i(Array, Id, I, NewId)
        ->
        id_type(Id,Tau), id_type(NewId,NewTau),
        arity(Tau, N), arity(NewTau, NewN),
        functor(Type, Tau, N), functor(NewType, NewTau, NewN),
        fold(Type, NewType, tclp__type_order:add_eq, !, Goal)
        ;
        Goal=fail.

make_bi_clause_array(Matrix, LUBMatrix, Array) :-
	make_lb_wrt_i_array(Matrix, LUBMatrix, IdArray),
	make_bi_clause_array(IdArray, Array).
	
make_bi_clause_array(IdArray, Array) :-
	current_counter_value(tclp_type_nb,Size),
	create_array(Size, Array),
	init_bi_clause_array(IdArray, Array, 1, Size).
init_bi_clause_array(_, _,N,Size) :- N > Size ,!.
init_bi_clause_array(IdArray, Array, N, Size) :-
	id_type(N,T), arity(T,A),
	functor(Term,T,A),
	array_element_at(Array,N,Term),
	init_bi_clause_array2(Term, 1, A, IdArray, N),
	N1 is N+1,
	init_bi_clause_array(IdArray, Array, N1, Size).
init_bi_clause_array2(_, I, A, _, _) :- I > A,!.
init_bi_clause_array2(Term, I, A, IdArray, N) :-
	arg(I,Term,g(Type, NewType, Goal)),
	make_arg_delete_clause_with_args(IdArray, N, I, Type, NewType, Goal),
	I1 is I+1,
	init_bi_clause_array2(Term, I1, A, IdArray, N).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% memoization of clauses
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% we use a micro CHR solver for keeping the data structure,
%% a global variable would have been better, and
%% assert won't work well because of the size of terms,
%% instead we just copy the needed part of the term

handler memo.

constraints memo/1.

consult @ memo(X) \ memo(Y) <=> X=Y.

%% build all order matrixes defined below 

build_matrixes :-
	verbose(debug_type_order,['build_matrixes ...']),
	create_order_matrix(Matrix),!,
	verbose(debug_type_order,['created order matrix']),
	reverse_matrix(Matrix,ReversedMatrix),!,
	verbose(debug_type_order,['created reversed matrix']),
	make_glb_matrix(Matrix,GLBMatrix),!,
	verbose(debug_type_order,['created glb matrix']),
	make_glb_matrix(ReversedMatrix,LUBMatrix),!,
	verbose(debug_type_order,['created lub matrix']),
	make_glb_clause_matrix(GLBMatrix, GLBClauseMatrix),!,
	verbose(debug_type_order,['created glb clause matrix']),
	make_lub_clause_matrix(LUBMatrix, LUBClauseMatrix),!,
	verbose(debug_type_order,['created lub clause matrix']),
	make_break_clause_matrix(Matrix, BreaKMatrix),!,
	verbose(debug_type_order,['created break matrix']),
	make_ordered_clause_matrix(Matrix, OrderedMatrix),!,
	verbose(debug_type_order,['created ordered matrix']),
	make_bi_clause_array(Matrix, LUBMatrix, LowArray),!,
	verbose(debug_type_order,['created low array']),
	make_bi_clause_array(ReversedMatrix, GLBMatrix, HiArray),!,
	verbose(debug_type_order,['created hi array']),
	memo(t(GLBClauseMatrix,
	       LUBClauseMatrix,
	       BreaKMatrix,
	       OrderedMatrix,
	       LowArray,
	       HiArray)),
	verbose(debug_type_order,['build_matrixes finished']).

%% acces clauses
get_glb_matrix(GLBClauseMatrix) :-
	memo(t(GLBClauseMatrix, _LUBClauseMatrix, _BreaKMatrix,
	       _OrderedMatrix, _LowArray, _HiArray)).
get_lub_matrix(LUBClauseMatrix) :-
	memo(t(_GLBClauseMatrix, LUBClauseMatrix, _BreaKMatrix,
	       _OrderedMatrix, _LowArray, _HiArray)).
get_break_matrix(BreaKMatrix) :-
	memo(t(_GLBClauseMatrix, _LUBClauseMatrix, BreaKMatrix,
	       _OrderedMatrix, _LowArray, _HiArray)).
get_ordered_matrix(OrderedMatrix) :-
	memo(t(_GLBClauseMatrix, _LUBClauseMatrix, _BreaKMatrix,
	       OrderedMatrix, _LowArray, _HiArray)).
get_lower_array(LowArray) :-
	memo(t(_GLBClauseMatrix, _LUBClauseMatrix, _BreaKMatrix,
	       _OrderedMatrix, LowArray, _HiArray)).
get_higher_array(HiArray) :-
	memo(t(_GLBClauseMatrix, _LUBClauseMatrix, _BreaKMatrix,
	       _OrderedMatrix, _LowArray, HiArray)).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% exported clauses
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

tclp__glb(Type1, Type2, TypeGLB) :-
	functor(Type1, Tau1,_), functor(Type2, Tau2,_),
	type_id(Tau1, Id1), type_id(Tau2, Id2),
	get_glb_matrix(GLBMatrix),
	matrix_element_at(GLBMatrix, Id1, Id2, G),
	copy_term(G,g(Type1, Type2, TypeGLB, Goal)),
	call(Goal),!.
tclp__glb(A,B,_) :-
	verbose(debug_type_order,['glb failed for ', A,' and ',B]),fail.

tclp__lub(B,T,T) :- builtin_type(bot,Bot),B==Bot,!.
tclp__lub(T,B,T) :- builtin_type(bot,Bot),B==Bot,!.
tclp__lub(Type1, Type2, TypeLUB) :-
	functor(Type1, Tau1,_), functor(Type2, Tau2,_),
	type_id(Tau1, Id1), type_id(Tau2, Id2),
	get_lub_matrix(LUBMatrix),
	matrix_element_at(LUBMatrix, Id1, Id2, G),
	copy_term(G,g(Type1, Type2, TypeLUB, Goal)),
	call(Goal),!.
tclp__lub(A,B,_) :-
	verbose(debug_type_order,['lub failed for ', A,' and ',B]),fail.

tclp__break(B,_) :-
	builtin_type(bot, Bot), B==Bot,!.
tclp__break(Type1, Type2) :-
	functor(Type1, Tau1,_), functor(Type2, Tau2,_),
	type_id(Tau1, Id1), type_id(Tau2, Id2),
	get_break_matrix(BreakMatrix),
	matrix_element_at(BreakMatrix, Id1, Id2, G),
	copy_term(G,g(Type1, Type2, Goal)),
	call(Goal),!.
tclp__break(A,B) :-
	verbose(debug_type_order,['break failed for ', A,' and ',B]),fail.

tclp__ordered_types(B,_) :-
	builtin_type(bot, Bot), B==Bot,!.
tclp__ordered_types(Type1, Type2) :-
	functor(Type1, Tau1,_), functor(Type2, Tau2,_),
	type_id(Tau1, Id1), type_id(Tau2, Id2),
	get_ordered_matrix(OrderedMatrix),
	matrix_element_at(OrderedMatrix, Id1, Id2, G),
	copy_term(G,g(Type1, Type2, Goal)),
	call(Goal),!.

tclp__elim_down_i(Type, I, NewType) :-
	functor(Type, Tau, _),
	type_id(Tau,Id),
	get_lower_array(Array), array_element_at(Array, Id, Term),
	arg(I, Term, G),
	copy_term(G,g(Type, NewType, Goal)),
	call(Goal),!.
tclp__elim_down_i(T,I,_) :-
	verbose(debug_type_order,['elim down failed for ', T,' and ',I]),fail.

tclp__elim_up_i(Type, I, NewType) :-
	functor(Type, Tau, _),
	type_id(Tau,Id),
	get_higher_array(Array), array_element_at(Array, Id, Term),
	arg(I, Term, G),
	copy_term(G,g(Type, NewType, Goal)),
	call(Goal),!.
tclp__elim_up_i(T,I,_) :-
	verbose(debug_type_order,['elim up failed for ', T,' and ',I]),fail.


%% type_exists(type)
%% checks if a type is well defined (all its components must have been
%% declared)
type_exists(X) :- var(X),!.
type_exists(Type) :-
        functor(Type,Head,NArgs),
	type_dec(Head, NArgs),
        type_exists(Type,NArgs).
type_exists(_,0) :- !.
type_exists(Type,N) :-
        arg(N,Type,Arg),
        type_exists(Arg),
        N1 is N-1,
        type_exists(Type,N1).
